package httpd;

import contrib.NanoHTTPD;
import sleep.runtime.*;
import sleep.bridges.*;
import sleep.interfaces.*;
import sleep.error.*;

import sleep.bridges.io.*;
import sleep.taint.TaintUtils;

import java.text.*;

import java.util.*;
import java.io.*;

import template.*;

public class Moconti extends NanoHTTPD implements RuntimeWarningWatcher, Loadable
{
   protected Map sites = new HashMap();
   protected ScriptLoader loader      = new ScriptLoader();
   protected Hashtable environment    = new Hashtable();
   protected Hashtable virtualContent = new Hashtable();
   protected String defaultSite;

   private static DateFormat logstamp = new SimpleDateFormat("dd/MMM/yyyy:HH:mm:ss");

   public class Session
   {
      protected ScriptInstance  theScript;
      protected File            rootDirectory;
      protected Map             hooks;
      protected long            lastUsed;
      protected String          id;
      protected boolean         needsInitialization;
      protected Site            parent;

      public String getId()
      {
         return id.toString();
      }

      /** check if the underlying source code powering this session has changed at all */
      public boolean hasChanged()
      {
         return theScript.hasChanged();
      }

      /** returns site lock */
      public sleep.bridges.Semaphore getSiteLock()
      {
         return parent.siteLock;
      }

      /** sets the shared data for these sessions */
      public void setSharedData(Object o)
      {
         parent.shared = o;
      }

      /** returns data structure for vars shared between sessions */
      public Object getSharedData()
      {
         return parent.shared;
      }

      /** register the specified URL with this session handler, eh?!? */
      public void addHook(String name, SleepClosure closure)
      {
         hooks.put(name, closure);
      }

      public Session(Site _parent, ScriptInstance script, File root, String key)
      {
         rootDirectory = root;
         theScript     = script;
         parent        = _parent;
         id            = key;

         /* setup some initial values */
         hooks = Collections.synchronizedMap(new HashMap());
         lastUsed = System.currentTimeMillis();

         /* init the id (if one wasn't provided) */
         if (id == null)
         {
            id = UUID.randomUUID().toString();
         }

         /* install the session into the script environment */
         theScript.getScriptVariables().putScalar("$session", SleepUtils.getScalar(this));

         /* init the session */
         theScript.runScript();

         /* flag this */
         needsInitialization = true;
      }
  
      public Response serve(String uri, String method, Properties header, Properties parms)
      {
         /* handle the actual request. */
         long start = System.currentTimeMillis();
         
         /* fix the URL to make / reflect index.slp or index.html */
         if (uri.endsWith("/") && hooks.get(uri) == null)
         {
            if (new File(new File(rootDirectory, uri), "index.slp").exists())
            {
               uri = uri + "index.slp";
            }
            else if (new File(new File(rootDirectory, uri), "index.html").exists())
            {
               uri = uri + "index.html";
            }
         }

         /* check if the URL is hooked or if it is a Sleep file */
         if (hooks.get(uri) != null || (uri.endsWith(".slp") && new File(rootDirectory, uri).exists()))
         {
            /* we always synchronize on the script variables for a script instance in a multi-threaded app.
               in the case of this webserver, multiple requests may happen at once, we sync here to allow the 
               setup, execution, and teardown to happen for each session. */
            synchronized (theScript.getScriptVariables())
            {
               /* check if the session is initialized or not, if its not then do something about it. */
               if (needsInitialization)
               {
                  /* now dump this initialized session into the cache, we want unused sessions to die
                     as quickly as possible--hence waiting to do all this */
                  parent.sessions.put(getId(), this);

                  /* can't forget to do this */
                  needsInitialization = false;
               }

               /* dump the request information to global vars, I chose global vars because this info
                  will be overwritten request to request anyways and it simplifies the web apps, makes
                  it so they don't have to pass this information around */

               Variable globals = theScript.getScriptVariables().getGlobalVariables();
               globals.putScalar("%header", TaintUtils.taint(SleepUtils.getHashWrapper(header)));
               globals.putScalar("%parms", TaintUtils.taint(SleepUtils.getHashWrapper(parms)));
               globals.putScalar("$method", TaintUtils.taint(SleepUtils.getScalar(method)));

               /* make STDOUT into a Sleep buffer object that we pass back to the client (eventually) */

               BufferObject buffer = new BufferObject();
               buffer.allocate(2048);
               IOObject.setConsole(theScript.getScriptEnvironment(), buffer);

               /* ok, now we execute the page */
              
               Scalar directives = null;

               if (hooks.get(uri) != null)
               {
                  /* call the closure */
                  SleepClosure closure = (SleepClosure)hooks.get(uri);
                  directives = SleepUtils.runCode(closure, uri, theScript, null); 
               }
               else
               {
                  /* execute .slp file */
                  try
                  { 
                     Stack locals = new Stack();
                     locals.push(  SleepUtils.getScalar(new File(rootDirectory, uri)  )   );
                     tengine.evaluate(uri, theScript, locals);
                  } 
                  catch (RuntimeException ex)
                  {
                     if (ex.getCause() instanceof YourCodeSucksException)
                     {
                        ((YourCodeSucksException)ex.getCause()).printErrors(System.err);
                     }
                     else if (ex.getCause() != null)
                     {
                        ex.getCause().printStackTrace();
                     }
                     else
                     {
                        ex.printStackTrace();
                     }
                  }
               }

               String data = new String(((ByteArrayOutputStream)buffer.getSource()).toByteArray());
 
               /* allow the return value of the response to alter our headers */
               Response response = new Response(HTTP_OK, MIME_HTML, data);

               if (!SleepUtils.isEmptyScalar(directives) && directives.getHash() != null)
               {
                  Iterator eset = directives.getHash().getData().entrySet().iterator();
                  while (eset.hasNext())
                  {
                     Map.Entry next = (Map.Entry)eset.next();

                     if (next.getKey().equals("status"))
                     {
                        response.status = next.getValue().toString();
                     }
                     else if (next.getKey().equals("Content-type"))
                     {
                        response.mimeType = next.getValue().toString();
                     }
                     else
                     {
                        response.addHeader((String)next.getKey(), (String)next.getValue().toString());
                     }
                  }  
               }

               /* make sure this session isn't going away any time soon */
               touch();

               /* set the cookie and return this response */
               response.addHeader("Set-Cookie", "session=" + getId() + ";EXPIRES="+getSessionExpiration()+"; PATH=/");
               return response;
            } /* end synchronization on script variables */
         }
         else if (uri.equals("/profiler"))
         {
            synchronized (theScript.getScriptVariables())
            {
               StringBuffer output = new StringBuffer("<html><body><title>Script Profiler</title>");

               output.append("<table><tr><th>Time Spent (s)</th><th>Total Calls</th><th>Function Description</th></tr>\n");

               Iterator i = theScript.getProfilerStatistics().iterator();
               while (i.hasNext())
               {
                  ScriptInstance.ProfilerStatistic temp = (ScriptInstance.ProfilerStatistic)i.next();
                  output.append("<tr><td>" + (temp.ticks / 1000.0) + "</td><td>" + temp.calls + "</td><td>" + temp.functionName + "</td></tr>\n");
               }

               output.append("</table></body></html>");

               return new Response(HTTP_OK, MIME_HTML, output.toString());
            }
         }
         else
         {
            return serveFile( uri, header, rootDirectory, true );
         }
      }

      public String toString()
      {
         return id + " (last used: " + ((System.currentTimeMillis() - lastUsed) / 1000.0) + "s)";
      }

      public void kill()
      {
         lastUsed = 0L;
      }

      public boolean isExpired()
      {
         return (System.currentTimeMillis() - lastUsed) > (1000 * 60 * 60); // expire sessions after one hour...
      }

      public void touch()
      {
         lastUsed = System.currentTimeMillis();
      }
   }

   /** obtain a session */
   public Session getSession(Site site, String key)
   {
      Map sessions = site.sessions;

      /** resolve the session for the specified key */
      if (key == null || sessions.get(key) == null || ((Session)sessions.get(key)).isExpired())
      {
         /** load the script */
         ScriptInstance script = loader.loadScriptNoReference(site.scriptName, site.compiledScript, (Hashtable)(environment.clone()));
         script.associateFile(new File(site.scriptName));

         script.chdir(site.resourceDirectory);

         /** create the session */
         Session temp = new Session(site, script, site.rootDirectory, key);

         /* return the session */
         return temp;
      }
      else
      {
         return (Session)sessions.get(key);
      }
   }

   private static String cformat = "EEE, dd-MMM-yyyy HH:mm:ss 'GMT'";

   private String getSessionExpiration()
   {
      Date d = new Date();
      d.setTime(System.currentTimeMillis() + (1000L * 60L * 60L * 24L * 365L));
      return new SimpleDateFormat(cformat).format(d);
   }

   public static void testit()
   {
       System.out.println(extractSessionId("session=f76fd845-1015-4b58-87d4-24796ab030d9; _csuid=491cce4b02535e5d"));
       System.out.println(extractSessionId("_csuid=491cce4b02535e5d;session=f76fd845-1015-4b58-87d4-24796ab030d9;fdsjfjkdsfjf=fdjsfjkdsjk"));
       System.out.println(extractSessionId("_csuid=491cce4b02535e5d; session=f76fd845-1015-4b58-87d4-24796ab030d9;"));
       System.out.println(extractSessionId("_csuid=491cce4b02535e5d; session=f76fd845-1015-4b58-87d4-24796ab030d"));
       System.out.println(extractSessionId("_csuid=491cce4b02535e5d; session=f76fd845-1015-4b58-87d4-24796ab030d9"));
       System.out.println(extractSessionId("_csuid=491cce4b02535e5d; sess=f76fd845-1015-4b58-87d4-24796ab030d9;"));
       System.out.println(extractSessionId("_csuid=491cce4b02535e5d; session=f76fd845-1015-4b58-8"));
       System.out.println(extractSessionId("session=f76fd845-1015-4b58-8"));
       System.out.println(extractSessionId("session=f76fd845-1015-4b58-8;"));
       System.out.println(extractSessionId("session=f76fd845-1015-4b58-87d4-24796ab030d9;"));
       System.out.println(extractSessionId("session=f76fd845-1015-4b58-87d4-24796ab030d9"));
       System.out.println(extractSessionId("f76fd845-1015-4b58-8"));
       System.out.println(extractSessionId(""));
       System.out.println(extractSessionId(null));
       System.out.println(extractSessionId("session=1"));
   } 

   /* extract the session id from the cookie... do it right this time please */
   private static String extractSessionId(String cookie)
   {
      try
      {
         if (cookie == null || cookie.length() == 0)
         {
            return cookie;
         }

         int start = cookie.indexOf("session=");

         if (start > -1)
         {
            start     = start + "session=".length();
            int clen  = start + "cb97431f-71b8-461a-bed4-12af382ce3a5".length();

            if (clen < cookie.length())
            {
               return cookie.substring(start, clen);
            }
            else
            {
               return cookie.substring(start);
            }
         }
      }
      catch (Exception ex)
      {
         System.out.println("extractCookieId('" + cookie + "') - FAIL: " + ex);
         ex.printStackTrace();
      }
      return cookie;
   }

   public Response serve(String uri, String method, Properties header, Properties parms)
   {
      long memory = Runtime.getRuntime().freeMemory() + (Runtime.getRuntime().maxMemory() - Runtime.getRuntime().totalMemory());

      /* check if a virtual file exists... this is valuable for "modules" that choose to
         hook in certain static files */
      File f = (File)virtualContent.get(uri);

      if (f != null && f.exists())
      {
         return serveFile(f.getPath(), header, new File("/"), true);
      }

      /* figure out which host we are serving */

      String from = (String)header.get(header.containsKey("X-Forwarded-For") ? "X-Forwarded-For" : "REMOTE_ADDRESS");
      String host = (String)header.get(header.containsKey("X-Forwarded-Host") ? "X-Forwarded-Host" : "Host"); 

      if (sites.get(host) == null)
      {
         host = defaultSite;
      }

      Site site = (Site)sites.get(host);

      /* check memory, take drastic measures if necessary */

      if (memory < (384 * 1024 * 1024))
      {
         System.err.println("[Moconti] low memory, clearing sessions: " + memory);
         site.sessions = Collections.synchronizedMap(new SessionCache()); /* kill all the current sessions */
         System.gc();

         if (memory < (96 * 1024 * 1024))
         {
            System.err.println("[Moconti] turned a user away for low memory: " + memory);
            return new Response(HTTP_OK, MIME_PLAINTEXT, "\"I've giv'n her all she's got captain, an' I canna give her no more.\"\n\n        Server is currently under load, try again in a few minutes");
         }
      }

      /* obtain the session key please */     

      String key;
      
      if (site.key == null)
      {
         key = extractSessionId(  (String)header.get("Cookie")  );
      }
      else if (parms.containsKey(site.key))
      {
         key = parms.get(site.key) + "";
      }
      else
      {
         key = from;
      }

      Session session = getSession(site, key);

      /* refresh the script if it has changed */

      if (session.hasChanged())
      {
         System.err.println("[Moconti] " + site.scriptName + " changed, reloading");
         addSite(host, site.scriptName, site.rootDirectory.toString(), site.resourceDirectory.toString());

         /* redo our variables please */
         site = (Site)sites.get(host);
         session = getSession(site, key);
      }

      Response r = session.serve(uri, method, header, parms);

      if (site.access_log != null)
      {   
         /* generate log entry */
         StringBuffer entry = new StringBuffer(512);
         entry.append(from);
         entry.append(" - - [");
         entry.append(logstamp.format(new Date(System.currentTimeMillis())));
         entry.append("] \"");
         entry.append(method);    
         entry.append(" ");
         entry.append(uri); 
         entry.append(" HTTP/1.0\" ");
         entry.append(r.status.substring(0, 3));
         entry.append(" ");
         entry.append(r.size);
         entry.append(" \"");
         entry.append(header.containsKey("Referer") ? header.get("Referer") : "");
         entry.append("\" \"");
         entry.append(header.get("User-Agent"));
         entry.append("\"");

         /* write it out */
         site.access_log.println(entry);
      }

      return r;
   }

   protected TemplateEngine tengine;

   protected Moconti() throws Exception
   { 
      super(  Integer.parseInt( System.getProperty("server.port", "1048") ), false ); 
      loader.addSpecificBridge(this);
      tengine = new TemplateEngine();
      loader.addGlobalBridge(tengine);
   }

   public void scriptLoaded(ScriptInstance script)
   {
      script.addWarningWatcher(this);
      script.getScriptVariables().putScalar("$server", SleepUtils.getScalar(this));

      if (System.getProperty("sleep.debug") != null)
      {
         script.setDebugFlags(Integer.parseInt(System.getProperty("sleep.debug")));
      }
   }

   public void scriptUnloaded(ScriptInstance script)
   {
   }

   public void addStatic(String uri, String file)
   {
      if (!new File(file).exists())
      {
         throw new IllegalArgumentException("file " + file + " does not exist");
      }
   
      try
      {
         virtualContent.put(uri, new File(file).getCanonicalFile());
      }
      catch (Exception ex)
      {
         throw new RuntimeException(ex);
      }
   }

   public void addSite(String h, String s, String rd, String resd)
   {
      addSite(h, s, rd, resd, null);
   }

   public void addSite(String host, String script, String rootDirectory, String resourceDirectory, String key)
   {
      Site site = new Site();
      site.rootDirectory = new File(rootDirectory).getAbsoluteFile();
      site.scriptName = script;
      site.sessions = Collections.synchronizedMap(new SessionCache());
      site.resourceDirectory = new File(resourceDirectory == null ? "." : resourceDirectory).getAbsoluteFile(); 
      site.key = key;  

      try
      {
         new File("logs").mkdir();
         site.access_log = new PrintStream(new FileOutputStream(new File("logs", host + ".log"), true));
      }
      catch (Exception ex)
      {
         System.err.println("[Moconti] could not create log files: " + ex);
      }

      if (!new File(script).exists())
      {
         System.err.println("[Moconti] unable to register " + host + ": " + script + " does not exist");
         return;
      }

      try
      {
         site.compiledScript = loader.compileScript(script);
      }
      catch (IOException ioex)
      {
         System.err.println(ioex);
      }
      catch (YourCodeSucksException yex)
      {
         yex.printErrors(System.err);
      }     

      if (!site.rootDirectory.exists())
      {
         System.err.println("[Moconti] unable to register " + host + ": " + rootDirectory + " does not exist");
         return;
      }

      defaultSite = host;
      sites.put(host, site);

      getSession(site, "__INIT__");
   }

   public void processScriptWarning(ScriptWarning warning)
   {
      System.err.println((new Date()) + " " + warning.toString());
   }

   public static void main(String args[])
   {
      String script = args.length == 0 || !(new File(args[0])).exists() ? "moconti.sl" : args[0];
 
      try
      {
         Moconti server = new Moconti();
         server.loader.setGlobalCache(true);

         ScriptInstance config = server.loader.loadScript(script, server.environment);
         config.getScriptVariables().putScalar("$server", SleepUtils.getScalar(server));
         config.runScript();

         server.listen();

         while (true)
         {
            Thread.currentThread().sleep(1000 * 60);
         }
      }
      catch (YourCodeSucksException yex)
      {
         yex.printErrors(System.out);
      }
      catch (Exception ex)
      {
         ex.printStackTrace();
      }
   }
}
