package httpd;

import java.util.*;

public class SessionCache extends LinkedHashMap
{
   private static long lowMemoryThreshold = 256 * 1024 * 1024; /* 32mb */

   protected boolean removeEldestEntry(Map.Entry eldest)
   {
      Moconti.Session last = (Moconti.Session)eldest.getValue();
      long memory = Runtime.getRuntime().freeMemory() + (Runtime.getRuntime().maxMemory() - Runtime.getRuntime().totalMemory());

      if (last.isExpired())
      {
         int count = 0;

         Iterator i = entrySet().iterator();
         while (i.hasNext())
         {
            Map.Entry entry = (Map.Entry)i.next();
            Moconti.Session session = (Moconti.Session)entry.getValue();

            if (session.isExpired())
            {
               count++;
               i.remove();
            }
            else
            {
               break;
            }
         }

         //System.err.println("[Moconti] expired " + count + " session(s) - free memory is now: " + memory);
         //System.gc();

         return true;
      }
      else
      {
         if (memory < lowMemoryThreshold)
         {
            System.err.println("[Moconti] available memory ("+memory+") is below threshold (" + size() + " sessions)");
            System.gc();
            return true;
         }
      }

      return false;
   }

   public SessionCache()
   {
      super(89, 0.75f, true);
   }
}
