package template;

import sleep.runtime.*;
import sleep.bridges.*;
import sleep.bridges.io.*;
import sleep.engine.*;
import sleep.interfaces.*;

import java.io.*;
import java.util.*;

/** a template object ready for evaluation */
public class Template
{
   /** the time when this template was loaded */
   protected long loadedTime;

   /** the source file for this template */
   protected File sourceFile;

   /** the template data */
   protected TemplateData first;
   protected TemplateData last;

   /** construct an empty template */
   public Template(File sFile)
   {
      sourceFile = sFile;
      loadedTime = System.currentTimeMillis();
   }

   private static class TemplateData
   {
      public int          type;
      public Block        block;
      public String       text;
      public TemplateData next;
   }

   protected void runCode(ScriptEnvironment environment, Variable locals)
   {
      TemplateData data = first;
      IOObject console = IOObject.getConsole(environment);

      while (data != null)
      {
         if (data.type == 0)
         {
            console.print(data.text);
         }
         else if (data.type == 1)
         {
            data.block.evaluate(environment);
         }

         data = data.next;
      }
   }

   /** evaluates the template */
   public void evaluate(ScriptInstance script, Stack locals)
   {
       ScriptEnvironment environment = script.getScriptEnvironment();

       synchronized (environment.getScriptVariables())
       {
          environment.pushSource(sourceFile.getName());
          environment.CreateFrame(); /* this frame holds the result */
          environment.getScriptVariables().pushLocalLevel(); /* create some local variable action */

          /* install the local  vars */
          Variable localv = environment.getScriptVariables().getLocalVariables();
          BridgeUtilities.initLocalScope(environment.getScriptVariables(), localv, locals);

          /* run the actual code, plz */
          runCode(environment, localv);

          /* handle the cleanup */
          environment.getScriptVariables().popLocalLevel();
          environment.KillFrame();
          environment.popSource();

          /* necessary since we're doing this from a toplevel */
          environment.resetEnvironment();
       }
   }

   /** check if this template is still valid (otherwise it must be reloaded) */
   public boolean isValid()
   {
      return loadedTime > sourceFile.lastModified();
   }

   public void addString(String text)
   {
      TemplateData temp = new TemplateData();
      temp.type = 0;
      temp.text = text;

      add(temp);
   }

   protected void add(TemplateData n)
   {
      if (first == null)
      {
         first = n;
      }
      else
      {
         last.next = n;
      }

      last = n;
   }

   public void addStatements(Block block)
   {
      TemplateData temp = new TemplateData();
      temp.type = 1;
      temp.block = block;

      add(temp);
   }
}
